'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var rxjs = require('rxjs');
var isMatch = _interopDefault(require('lodash.ismatch'));
var operators = require('rxjs/operators');

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(n);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _createForOfIteratorHelperLoose(o) {
  var i = 0;

  if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
    if (Array.isArray(o) || (o = _unsupportedIterableToArray(o))) return function () {
      if (i >= o.length) return {
        done: true
      };
      return {
        done: false,
        value: o[i++]
      };
    };
    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  i = o[Symbol.iterator]();
  return i.next.bind(i);
}

// @ts-nocheck
// wrt to a returned Observable - one will be canceled if its running,
// one will only be started if it wasn't running!
//prettier-ignore

var toggleMap = function toggleMap(spawner, mapper) {
  if (mapper === void 0) {
    mapper = function mapper(_, inner) {
      return inner;
    };
  }

  return function (source) {
    return new rxjs.Observable(function (observer) {
      var innerSub;
      return source.subscribe({
        next: function next(outer) {
          if (!innerSub || innerSub.closed) {
            innerSub = spawner(outer).subscribe(function (inner) {
              return observer.next(mapper(outer, inner));
            }, function (e) {
              return observer.error(e);
            });
          } else {
            innerSub.unsubscribe();
          }
        },
        error: function error(e) {
          observer.error(e);
        },
        complete: function complete() {
          observer.complete();
        }
      });
    });
  };
};

(function (ConcurrencyMode) {
  /**
   * Newly returned Observables are subscribed immediately, without regard to resource constraints, or the ordering of their completion. (ala mergeMap) */
  ConcurrencyMode["parallel"] = "parallel";
  /**
   * Observables are enqueued and always complete in the order they were triggered. (ala concatMap)*/

  ConcurrencyMode["serial"] = "serial";
  /**
   * Any existing Observable is canceled, and a new is begun (ala switchMap) */

  ConcurrencyMode["replace"] = "replace";
  /**
   * Any new Observable is not subscribed if another is running. (ala exhaustMap) */

  ConcurrencyMode["ignore"] = "ignore";
  /**
   * Any new Observable is not subscribed if another is running, and
   * the previous one is canceled. (ala switchMap with empty() aka toggleMap) */

  ConcurrencyMode["toggle"] = "toggle";
})(exports.ConcurrencyMode || (exports.ConcurrencyMode = {}));

var Channel = /*#__PURE__*/function () {
  function Channel() {
    this.channel = new rxjs.Subject();
    this.filters = new Map();
    this.listeners = new Map();
    this.listenerEnders = new Map();
    this.listenerParts = new Map();
    this.errors = new rxjs.Subject();
    this.logs = new rxjs.Subject();

    if (!isTestMode()) {
      this.errors.subscribe(function (e) {
        return console.error(e);
      });
      this.logs.subscribe(function (args) {
        var _console;

        return (_console = console).log.apply(_console, args);
      });
    }
  }

  var _proto = Channel.prototype;

  _proto.trigger = function trigger(eventOrType, payload, config) {
    var event = typeof eventOrType === 'string' ? {
      type: eventOrType
    } : eventOrType;
    payload && Object.assign(event, {
      payload: payload
    });

    for (var _iterator = _createForOfIteratorHelperLoose(this.filters.entries()), _step; !(_step = _iterator()).done;) {
      var _step$value = _step.value,
          predicate = _step$value[0],
          _filter2 = _step$value[1];
      predicate(event) && _filter2(event);
    }

    if (config == null ? void 0 : config.result) {
      event.result = this.query(config.result).toPromise();
    }

    Object.freeze(event);
    this.channel.next(event);

    for (var _iterator2 = _createForOfIteratorHelperLoose(this.listeners.entries()), _step2; !(_step2 = _iterator2()).done;) {
      var _step2$value = _step2.value,
          _predicate = _step2$value[0],
          listener = _step2$value[1];

      if (_predicate(event)) {
        listener(event);
      }
    }

    return event;
  }
  /**
   * Provides an Observable of matching events from the channel.
   */
  ;

  _proto.query = function query(eventMatcher, payloadMatcher) {
    var resultObs = this.channel.asObservable().pipe(operators.filter(getEventPredicate(eventMatcher, payloadMatcher)), operators.map(function (e) {
      return e;
    })); // @ts-ignore

    resultObs.toPromise = function () {
      return resultObs.pipe(operators.first()).toPromise();
    }; // @ts-ignore


    resultObs.then = function (resolve, reject) {
      return resultObs.toPromise().then(resolve, reject);
    };

    return resultObs;
  };

  _proto.filter = function filter(eventMatcher, f) {
    var _this = this;

    var predicate = getEventPredicate(eventMatcher);
    this.filters.set(predicate, f);
    return new rxjs.Subscription(function () {
      _this.filters["delete"](predicate);
    });
  };

  _proto.listen = function listen(eventMatcher, listener, config) {
    var _this2 = this;

    if (config === void 0) {
      config = {};
    }

    var predicate = getEventPredicate(eventMatcher);
    var userTriggers = config.trigger || {};
    var ender = new rxjs.Subject();
    var parts = new rxjs.Subject();
    this.listenerEnders.set(predicate, ender);
    this.listenerParts.set(predicate, parts);
    var canceler = new rxjs.Subscription(function () {
      _this2.deactivateListener(predicate);
    });

    var enqueuePart = function enqueuePart(event) {
      try {
        // @ts-ignore
        var userReturned = toObservable(listener(event));
        var base = rxjs.concat(userReturned, applyCompleteTrigger());
        var part = config.takeUntil ? base.pipe(operators.takeUntil(_this2.query(config.takeUntil))) : base;
        parts.next(part);
      } catch (e) {
        _this2.errors.next(e);

        _this2.errors.next("A listener function threw an exception and will be unsubscribed");

        canceler.unsubscribe();
      }
    };

    var applyNextTrigger = function applyNextTrigger(e) {
      //@ts-ignore
      userTriggers.next && _this2.trigger(userTriggers.next, e);
      userTriggers === true && _this2.trigger(e);
    };

    var applyCompleteTrigger = function applyCompleteTrigger() {
      return new rxjs.Observable(function (notify) {
        //@ts-ignore
        userTriggers.complete && _this2.trigger(userTriggers.complete);
        notify.complete();
      });
    };

    var applyOverlap = operatorForMode(exports.ConcurrencyMode[config.mode || 'parallel']);
    var listenerEvents = parts.pipe(applyOverlap(function (part) {
      return part;
    }), operators.tap(applyNextTrigger), operators.takeUntil(ender));
    var listenerObserver = {
      error: function error(err) {
        _this2.errors.next(err);

        _this2.errors.next("A listener function notified with an error and will be unsubscribed");

        canceler.unsubscribe(); // @ts-ignore

        userTriggers.error && _this2.trigger(userTriggers.error, err);
      }
    };
    var listenerSub = listenerEvents.subscribe(listenerObserver);
    canceler.add(function () {
      return listenerSub.unsubscribe();
    });
    this.listeners.set(predicate, enqueuePart);
    return canceler;
  };

  _proto.on = function on(eventMatcher, listener, config) {
    if (config === void 0) {
      config = {};
    }

    return listen(eventMatcher, listener, config);
  };

  _proto.spy = function spy(listener) {
    var _this3 = this;

    var sub;

    var safelyCallListener = function safelyCallListener(event) {
      try {
        return listener(event);
      } catch (err) {
        _this3.errors.next(err);

        _this3.errors.next("A listener function threw an exception and will be unsubscribed");

        if (sub) {
          sub.unsubscribe();
        }
      }
    };

    sub = this.filter(true, safelyCallListener);
    return sub;
  }
  /** Logs to the console (but not in test mode) */
  ;

  _proto.log = function log() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    this.logs.next(args);
  };

  _proto.reset = function reset() {
    this.filters.clear();

    for (var _iterator3 = _createForOfIteratorHelperLoose(this.listeners.keys()), _step3; !(_step3 = _iterator3()).done;) {
      var listenerPredicate = _step3.value;
      this.deactivateListener(listenerPredicate);
    }
  };

  _proto.deactivateListener = function deactivateListener(predicate) {
    // unregister from future effects
    this.listeners["delete"](predicate); // cancel any in-flight

    var ender = this.listenerEnders.get(predicate);
    ender && ender.next();
    this.listenerEnders["delete"](predicate);
  };

  return Channel;
}(); // Exports for a default Channel

var channel = /*#__PURE__*/new Channel();
var trigger = /*#__PURE__*/channel.trigger.bind(channel);
var query = /*#__PURE__*/channel.query.bind(channel);
var filter = /*#__PURE__*/channel.filter.bind(channel);
var listen = /*#__PURE__*/channel.listen.bind(channel);
var on = /*#__PURE__*/channel.on.bind(channel);
var spy = /*#__PURE__*/channel.spy.bind(channel);
var log = /*#__PURE__*/channel.log.bind(channel);
var reset = /*#__PURE__*/channel.reset.bind(channel);

function isTestMode() {
  var _process, _process$env;

  if (typeof process === 'undefined') return false;
  return ((_process = process) == null ? void 0 : (_process$env = _process.env) == null ? void 0 : _process$env.NODE_ENV) === 'test';
}

function getEventPredicate(eventMatcher, payloadMatcher) {
  var predicate;

  if (eventMatcher instanceof RegExp) {
    predicate = function predicate(event) {
      return eventMatcher.test(event.type) && (!payloadMatcher || isMatch(event.payload, payloadMatcher));
    };
  } else if (eventMatcher instanceof Function) {
    predicate = eventMatcher;
  } else if (typeof eventMatcher === 'boolean') {
    predicate = function predicate() {
      return eventMatcher;
    };
  } else if (typeof eventMatcher === 'object') {
    predicate = function predicate(event) {
      return isMatch(event, eventMatcher);
    };
  } else if (eventMatcher.constructor === Array) {
    predicate = function predicate(event) {
      return eventMatcher.includes(event.type) && (!payloadMatcher || isMatch(event.payload, payloadMatcher));
    };
  } else {
    predicate = function predicate(event) {
      return eventMatcher === event.type && (!payloadMatcher || isMatch(event.payload, payloadMatcher));
    };
  }

  return predicate;
}
/** Controls what types can be returned from an `on` handler:
    Primitive types: `of()`
    Promises: `from()`
    Observables: pass-through
*/


function toObservable(_results) {
  if (typeof _results === 'undefined') return rxjs.of(undefined);

  if (typeof _results === 'function') {
    return _results.length === 1 ? new rxjs.Observable(_results) : rxjs.defer(_results);
  } // An Observable is preferred


  if (_results.subscribe) return _results; // A Subscrition is ok - can be canceled but not awaited

  if (_results.unsubscribe) return new rxjs.Observable(function () {
    // an Observable's return value is its cleanup function
    return function () {
      return _results.unsubscribe();
    };
  }); // A Promise or generator is acceptable

  if (_results.then || typeof _results[Symbol.iterator] === 'function') return rxjs.from(_results); // otherwise we convert it to a single-item Observable

  return rxjs.of(_results);
}

function operatorForMode(mode) {
  if (mode === void 0) {
    mode = exports.ConcurrencyMode.parallel;
  }

  switch (mode) {
    case exports.ConcurrencyMode.ignore:
      return operators.exhaustMap;

    case exports.ConcurrencyMode.parallel:
      return operators.mergeMap;

    case exports.ConcurrencyMode.serial:
      return operators.concatMap;

    case exports.ConcurrencyMode.toggle:
      return toggleMap;

    case exports.ConcurrencyMode.replace:
    default:
      return operators.switchMap;
  }
}

/**
 * Returns a random hex string, like a Git SHA. Not guaranteed to
 * be unique - just to within about 1 in 10,000.
 */

var randomId = function randomId(length) {
  if (length === void 0) {
    length = 7;
  }

  return Math.floor(Math.pow(2, length * 4) * Math.random()).toString(16).padStart(length, '0');
};
/**
 * Returns an Observable of the value, or result of the function call, after
 * the number of milliseconds given. After is lazy and cancelable! So nothing happens until .subscribe
 * is called explicitly (via subscribe) or implicitly (toPromise(), await).
 * For a delay of 0, the function is executed synchronously when .subscribe is called.
 * @returns An Observable of the object or thunk return value. It is 'thenable', so may also be awaited directly.
 * ```
 * // Examples:
 * // awaited Promise
 * await after(100, () => new Date())
 * // unawaited Promise
 * after(100, () => new Date()).toPromise()
 * // unresolving Promise
 * after(Infinity, () => new Date()).toPromise()
 * ```
 */

function after(ms, objOrFn) {
  var delay = ms <= 0 ? rxjs.of(0) : ms === Infinity ? rxjs.NEVER : rxjs.timer(ms);
  var resultObs = delay.pipe(operators.map(function () {
    return typeof objOrFn === 'function' ? objOrFn() : objOrFn;
  })); // after is a 'thenable, thus usable with await.
  // ref: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/await
  // @ts-ignore

  resultObs.then = function (resolve, reject) {
    return resultObs.toPromise().then(resolve, reject);
  }; // @ts-ignore


  return resultObs;
}
/** Executes the given function on the microtask queue.
 * The microtask queue flushes before the macrotask queue.
 * @returns A Promise for its return value
 * @see https://developer.mozilla.org/en-US/docs/Web/API/HTML_DOM_API/Microtask_guide
 * @see queueMicrotask, Promise.resolve
 */

function microq(fn) {
  // @ts-ignore
  return Promise.resolve().then(fn);
}
/** Executes the given function on the macrotask queue.
 * The macrotask queue flushes after the microstask queue.
 * @returns A Promise for its return value
 * @see https://developer.mozilla.org/en-US/docs/Web/API/HTML_DOM_API/Microtask_guide
 * @see setTimeout
 */

function macroq(fn) {
  return new Promise(function (resolve) {
    return setTimeout(function () {
      return resolve(fn());
    }, 0);
  });
}

var getTimestamp = function getTimestamp() {
  return new Date().getTime();
};
/** Returns a Promise for the point in time at which all existing queued microtasks (e.g. Promise.resolve()) have completed. */


function microflush() {
  return Promise.resolve().then(function () {
    return getTimestamp();
  });
}
/** Returns a Promise for the point in time at which all existing queued macrotasks (e.g. setTimeout 0) have completed. */

function macroflush() {
  return new Promise(function (resolve) {
    return setTimeout(function () {
      return resolve(getTimestamp());
    }, 0);
  });
}
/** Decorates a function so that its argument is the mutable array
 * of all events seen during its run. Useful for testing:
 *
 * it('does awesome', captureEvents(async seen => {
 *  trigger('foo)
 *  expect(seen).toEqual([{type: 'foo'}])
 * }));
 */

function captureEvents(testFn) {
  return function () {
    var seen = new Array(); // @ts-ignore

    var sub = query(true).subscribe(function (event) {
      return seen.push(event);
    });
    var result = testFn(seen);

    if (result && result.then) {
      return result["finally"](function () {
        return sub.unsubscribe();
      });
    }

    sub.unsubscribe();
    return result;
  };
}

Object.defineProperty(exports, 'Subscription', {
  enumerable: true,
  get: function () {
    return rxjs.Subscription;
  }
});
Object.defineProperty(exports, 'concat', {
  enumerable: true,
  get: function () {
    return rxjs.concat;
  }
});
Object.defineProperty(exports, 'map', {
  enumerable: true,
  get: function () {
    return operators.map;
  }
});
Object.defineProperty(exports, 'scan', {
  enumerable: true,
  get: function () {
    return operators.scan;
  }
});
Object.defineProperty(exports, 'tap', {
  enumerable: true,
  get: function () {
    return operators.tap;
  }
});
exports.Channel = Channel;
exports.after = after;
exports.captureEvents = captureEvents;
exports.channel = channel;
exports.filter = filter;
exports.listen = listen;
exports.log = log;
exports.macroflush = macroflush;
exports.macroq = macroq;
exports.microflush = microflush;
exports.microq = microq;
exports.on = on;
exports.query = query;
exports.randomId = randomId;
exports.reset = reset;
exports.spy = spy;
exports.toggleMap = toggleMap;
exports.trigger = trigger;
//# sourceMappingURL=polyrhythm.cjs.development.js.map
